#Relative imports
from .random import random_complex;from .basic import isNumber,isInteger;from .basic import product; 
#Built-in imports
from typing import Union
import re

NUMBER_REGEX = r"(((\-)|(\+))?\d+((\.)\d+)?)"

class Polynomial:
    def __init__(self,coefficients : list):
        degrees = {}
        i = 0
        coefficients = EnsureDegree(coefficients)
        for coefficient in coefficients:
            degrees[i] = coefficient
            i+=1
        if coefficients.count(0) == len(coefficients):
            degrees = {0 : 0}
        if len(coefficients) == 1:
            degrees = {0 : coefficients[0]}
        self.degree = i-1
        self.equation = degrees
        self.array = coefficients
        self.function = eval("lambda x :" + self.__str__(useSymbol=True).split(":")[1].strip().replace("^","**"))
        self.type = type(self)

    def eq(self):
        return self.equation

    def arr(self):
        return self.array

    def getFunction(self):
        return self.function

    def deg(self):
        return self.degree

    def __str__(self,useSymbol : bool = False):
        if useSymbol:
            return self.__old_str__(useSymbol=True)

        eq = []
        j = 0
        for item in self.equation:
            if self.equation.get(item) == 0:
                continue
            val = self.equation.get(item)
            sign = f'{"+" if item != len(self.equation)-1 and not "-" in str(self.equation.get(item)) else ""}'
           
            if type(val) != complex and isInteger(val): #Prevent the rise of an Exception
                target = str(int(val)) #Display Integer Value not rounded
            else:
                target = str(val)  #Display vloat value

            if type(val) != complex: #Put Spaced between the negative sign if Not in the begining
                target = target.replace("-","- ")
            
            if re.sub(r"\s+","",target) in ("-1","1") and item !=0: #Remove the 1 if it is not a constant
                target = target.replace("1","")
            
            use_pow = "x^" + str(item) if item not in (0,1) else "x" if item == 1  else "" #handle exponentiation
            eq.append(f'{sign} {target}{"*" if useSymbol else ""}{use_pow} ')
            j+=1
        Joined = "".join(list(reversed(eq))).strip()
        if Joined.strip()[-1] == "*":
            Joined = Joined[:-1]
        return f'Polynomial of degree {self.degree} : {Joined}'  
    

    def __old_str__(self,useSymbol): 
        """The unreadable old str method which needs to stay in there in order for the lamda getfunc to work"""
        eq = []
        j = 0
        for item in self.equation:  
            if self.equation.get(item) == 0:
                continue
            eq.append(f'{"+" if j != len(self.equation)-1 and not "-" in str(self.equation.get(item)) else ""  } {"(" if "j" in str(self.equation.get(item)) else ""}{str(self.equation.get(item)).replace("-","- ")}{")" if "j" in str(self.equation.get(item)) else ""}{"*" if useSymbol else ""}{"x^" + str(item) if item not in (0,1) else "x" if item == 1  else ""} ')
            j+=1
        Joined = "".join(list(reversed(eq))).strip()
        if Joined.strip()[-1] == "*":
            Joined = Joined[:-1]
        return f'Polynomial of degree {self.degree} : {Joined}'   

    __repr__ = __str__

    def diffrentiate(self,getFunction: bool = False):
        derivate = {}
        for term in self.equation:
            derivate[term-1] = (term)*self.equation.get(term)
        derivate.pop(-1)
        if not getFunction:
            return Polynomial([derivate.get(item) for item in derivate])
        return eval("lambda x : " + Polynomial(derivate).__str__(useSymbol=True).split(":")[1].strip().replace("^","**"))

    def integrate(self,getFunction: bool = False):
        integral = {}
        for term in self.equation:
            integral[term+1] = round(1 / (term+1),3) * self.equation.get(term) if term !=0 else self.equation.get(term)
        if not getFunction:
            for i in range(max(list(integral))):
                if i not in integral:
                    integral[i] = 0
            returntype = list(integral)
            returntype.sort()
            return Polynomial([integral.get(item) for item in returntype])

    def __neg__(self):
        return -1 * self

    def __add__(self,value : [int,float,complex,'Polynomial']) -> 'Polynomial':
        eq_copy = self.equation.copy()
        #Scalar
        if type(value) in (int,float,complex):
            eq_copy[0] = eq_copy[0] + value if 0 in eq_copy else value
            return Polynomial([eq_copy.get(item) for item in eq_copy])
        #Polynomial
        elif type(value) == type(self):
            array_1 = [value.eq().get(item) for item in value.eq()] #value items
            array_2 = [self.equation.get(item) for item in self.equation] #self items
            if len(value.eq()) > len(self.eq()): #Value Greater
                #Add to the greater
                for i in range(len(array_2)):
                    array_1[i] += array_2[i]
                return Polynomial(array_1)

            for i in range(len(array_1)): #Self is greater
                #Add to the greater
                array_2[i] += array_1[i]
            return checkPolynomial(array_2)
        raise TypeError("Cannot add Polynomial with {}".format(type(value)))

    def __radd__(self,value): #Right add
        """For adding something to a polynomial (int + Polynomial)"""
        return self.__add__(value)

    def __sub__(self,value : [int,float,complex,'Polynomial']) -> Union['Polynomial',float]:
        eq_copy = self.equation.copy()
        #Scalar
        if type(value) in (int,float,complex):
            eq_copy[0] = eq_copy[0] - value if 0 in eq_copy else value
            return Polynomial([eq_copy.get(item) for item in eq_copy])
        elif type(value) == type(self):
            return self + (-1 * value)
        raise TypeError("Cannot perform subtraction on Polynomial with {}".format(type(value)))

    def __rsub__(self,value): #Right sub
        return -self + value

    def __mul__(self,value : [int,float,complex,"Polynomial"]) -> Union['Polynomial',float]:
        if type(value) in (int,float,complex):
            return checkPolynomial([value * num for num in self.array])
        
        elif type(value) == self.type:
            dick0 = self.equation
            dick1 = value.eq()
            new_dick = {}
            for item in dick0:
                for value in dick1:
                    if not (value+item) in new_dick:
                        new_dick[value+item] = []
                    new_dick[value+item].append(dick1.get(value) * dick0.get(item))
            for num in new_dick:
                new_dick[num] = sum(new_dick[num])
            item_max = max([item for item in new_dick])
            for i in range(item_max):
                if i not in new_dick:
                    new_dick[i] = 0
            x = [item for item in new_dick]
            x.sort()
            return checkPolynomial([new_dick.get(item) for item in x])

        return NotImplemented #You can redefine it with __rmul__

    def __rmul__(self,value : [int,float,complex,"Polynomial"]) -> Union['Polynomial',float]: #Right mul
        return self.__mul__(value)

    def __pow__(self,value : int) -> Union['Polynomial',float]:
        if not type(value) == int:
            raise TypeError("Cannot perform polynomial exponentiation with {}".format(type(value)))
        return product(*[self for i in range(value)])
    
    def __truediv__(self,value : [int,float,complex,"Polynomial"]) -> Union[float,"Polynomial"]:
        """
            Handling division by a scalar or by a Polynomial

            # NOTE 1 :  Divide the term of the highest degree of the divisor (x^2)\n
            # NOTE 2 :  with the highest term of the number you are dividing (x)\n
            # NOTE 3 :  Mutliply the above result with the number you are dividing  x*(x+1)\n
            # NOTE 4 :  Now subtract from the divisor the above result x^2+2x+1 - (x^2 + x)\n
            # NOTE 5 :  Repeat the same thing with the above result\n
            # NOTE 6 :  Keep repeating the algorithm until you are left with a constant remainder\n

        """

        if type(value) in (int,float,complex): #Scalar division
            new_dict = self.equation.copy() 
            for item in new_dict: #Divide every single term with the scalar
                new_dict[item] = new_dict[item] / value 
            return checkPolynomial([new_dict.get(item) for item in new_dict])

        elif type(value) == self.type: #Polynomial division
            RESULT_DICT = {}
            if self.degree < value.deg():
                raise ValueError("Cannot divide Polynomial of degree {} with one of {} (The first polynomial must have a higher degree ({} < {})  )".format(self.degree,value.deg(),self.degree,value.deg()))
            self_copy = Polynomial(self.array.copy())
            value_copy = Polynomial(value.arr().copy())
            
            while type(self_copy) not in (int,float) and self_copy.deg() >= value_copy.deg():
                max_divisor_pow = max([num for num in self_copy.eq()]) #The highest power    (P1)
                max_dividand_pow = max([num for num in value_copy.eq()]) #The highest power      (P2)
                
                max_divisor = self_copy.eq().get(max_divisor_pow) #P1 coefficient
                max_dividand = value_copy.eq().get(max_dividand_pow)   #P2 coefficient


                div = max_divisor_pow - max_dividand_pow #'Devide' the degrees (by subtracting)
                div_const = max_divisor / max_dividand #Divide the constants


                new_poly_dict = PolynomialFromDict({div : div_const}) #The division result
                RESULT_DICT[div] = div_const


                times = new_poly_dict * value #The multiplication result
                self_copy -= times
                

            return [PolynomialFromDict(RESULT_DICT),self_copy] #Polynomial with reminder
        raise TypeError("Cannot divide Polynomial with {}".format(type(value)))

    def roots(self,iterations : int) -> complex:
        """It returns an list of complex numbers that are it roots or are approximately very close to them
           WARNING : THIS FUNCTION IS NOT STABLE AND REQUIRES ADJUSTING (OVERFLOW ERRORS DUE TO VERY LOW NUMBERS ARE COMMON)
        """
        reduced_pol = reduceCoefficients(self) #Make the leading coefficient 0
        result = applyKruger(reduced_pol.getFunction(),self.degree,iterations)
        return result

def kerner_durand(APPROXIMATIONS,function):
    TEMP_STORAGE = []
    for value in APPROXIMATIONS:
        current = APPROXIMATIONS.get(value)
        nominator = function(current) 
        #Handle Dividing
        denominator =  [] #list of all the subtractions
        cop = APPROXIMATIONS.copy()
        cop.pop(value)
        for item in cop:
            denominator.append(current - cop.get(item)) 
        TEMP_STORAGE.append(current - nominator / product(*denominator))
    i = 0 
    for item in APPROXIMATIONS:
        APPROXIMATIONS[item] = TEMP_STORAGE[i]
        i+=1
    return APPROXIMATIONS

def checkPolynomial(pol_list : list):
    pol_list = EnsureDegree(pol_list)
    if len(pol_list) == pol_list.count(0) or len(pol_list) == 1:
        return pol_list[0]
    return Polynomial(pol_list)
    
def applyKruger(function : callable,degree : int,iterations : int):
    APPROXIMATIONS = {}
    #Get our starting points
    for i in range(0,degree):
        APPROXIMATIONS[i] =  random_complex() #Begin with random numbers

    INITIAL = APPROXIMATIONS.copy()
    try:
        for i in range(iterations):
            APPROXIMATIONS = kerner_durand(APPROXIMATIONS,function)
    except:
        print("Root Calclualtion Failed")
        APPROXIMATIONS = INITIAL
        applyKruger(function,degree,iterations)
    #for visual purposes
    return [APPROXIMATIONS.get(item) for item in APPROXIMATIONS]

def PolynomialFromDict(poly_dict : dict) -> Union[Polynomial,float]:
    deg_array = [item for item in poly_dict]

    for i in range(int(max(deg_array))):
        if i not in poly_dict:
            poly_dict[i] = 0 
    deg_array = [item for item in poly_dict]
    deg_array.sort()
    return checkPolynomial([poly_dict.get(x) for x in deg_array])

def PolString(eqstring : str) -> Polynomial:
    #Check at the end of the string for constants
    eqstring = " " + re.sub(r"\s+",'',eqstring.strip()) #Remove all white space
    eqdict = {}
    exp_iteral = NUMBER_REGEX + r"?(-?x)?(\^(" + NUMBER_REGEX + r"))?"
    res = list(re.finditer(exp_iteral,eqstring))
    res = [item.group() for item in res if item.group().strip() != ""]
    for item in res:
        item = item.strip()
        if not 'x' in item: #Constants
            exponent = 0
            base = float(item)
        else:
            if '^' in item: #Power
                new_str = [x for x in item.split("x^") if x.strip() != '']
                if not item.startswith("x"):
                    base =  float(new_str[0]) if new_str[0] !='-' else -1
                    exponent = float(new_str[1])  
                else:
                    new_str_new = [x for x in item.split("^") if x.strip() !='']
                    base = float(new_str_new[0]) if new_str_new[0] != 'x' else 1
                    exponent = float(new_str_new[1]) if new_str_new[1] !='x' else 1
            else: #No Power
                if item.startswith('x'):
                    base = 1
                else:
                    replacement = item.replace("x",'')
                    if replacement.strip() =='-':
                        base = -1
                    else:
                        base = float(item.replace("x",''))
                exponent = 1
        if not exponent in eqdict:
            eqdict[exponent] = []
        eqdict[exponent].append(base)
    for expo in eqdict:
        eqdict[expo] = sum(eqdict[expo])
    return PolynomialFromDict(eqdict)

def EnsureDegree(poly_list : list):
    # Polynomial([5,0,0,0,0,0]) x^2+2x+3
    new_list = list(reversed(poly_list))
    while new_list[0] == 0:
        new_list.pop(0)
        if len(new_list) == 0:
            return [0]
    return list(reversed(new_list))

def LCM_POL(*polynomials : Polynomial): 
    """
           The Least common multiple of n polynomials will produce another polynomial
           that can be devided equally by all these n polynomials
    """
    return product(*polynomials)

def reduceCoefficients(polynomial : Polynomial) -> Polynomial:
    """
        Initialize a Polynomial for the Kerner method:
        Make the leading coefficient 1
    """
    eq = polynomial.eq()
    h_coeff = eq.get(max(eq))
    return Polynomial([
        coefficient / h_coeff for coefficient in polynomial.arr()
    ])

x = PolString("x")

if __name__ == "__main__":
    pass


"""
    Module containing some basic Math operations-functions:
        ** product(*args) # Returns the product of specified numbers
        ** comaSplitNumber(num : int or float) #Returns a human readble string splitting a large number with commas
        ** isNumber(arg) -> bool #Returns a boolean on wheter an argument can be converted to a float
        ** isInteger(num) -> bool # Returns a boolean on wheter int(num) == float(num)
        ** isRoot(function,x1,x2) -> bool # Given a function and 2 points, returns a boolean on wheter there is a root in that interval
"""

from typing import Union,Any,Tuple

def product(*args : Union[float,int]) -> Union[float,int]:
    """Returns the product of float or ints
        product(3,4,5) -> 60
        product(*[3,4,5]) -> 60
    """
    prod = 1
    for num in args:
        prod*=num
    return prod

def comaSplitNumber(num : Union[int,float]) -> str:
    """Given an integer or a floating number it returns
       a human readable string splitting the digits with commas
       example :
       input(1000) -> 1,000
       input(10000) -> 10,000
       input(1000.324) -> 1,000.324
    """
    num = str(num)
    oper_num = num.split('.')
    DIGITS = []
    iteration = 1
    for digit in reversed(oper_num[0]):
        DIGITS.append(digit)
        if(iteration%3==0):
            DIGITS.append(',')
        iteration+=1
    return_type = "".join(reversed(DIGITS))
    if(return_type.startswith(',')):
        return_type = return_type[1:len(return_type)]
    if (len(oper_num) > 1):
        return_type+=return_type[-1]
    return return_type

def isNumber(arg) -> bool:
    """Returns True if the value can be converted to a floating point"""
    try:
        int(arg)
        return True
    except:
        return False

def isInteger(num : Union[float,int]) -> bool:
    """Returns True if the int(num) is equal to float(num)\n
       it only handles numbers and will not care about other types
    """
    if(int(num)==float(num)):
        return True
    return False

def isComplex(num : Any) -> bool:
    """Returns a boolean on wheter or not a passed in argument really is a complex number
    """
    try:
        num_complex : complex = complex(num)
        return num_complex.real != num_complex
    except:
        return False


def Number(x : Any) -> Union[None,Tuple[int,float,complex]]:
    try:
        item = complex(x)
        if item.imag == 0:
            return item.real
        return item
    except:
        return None


def isRoot(function : callable,x_0 : float,x_1 : float) -> bool:
    """
        Given a function and 2 points in the x axis,\n
        it returns a boolean on wheter there is a root,\n
        in that interval
    """
    return function(x_0)*function(x_1)

if __name__ == "__main__":
    print(Number("5j"))


"""
    Module containing functions involed in number theory:
        ** isEven(x) x%2==0
        ** isOdd(x) x%2!=0
        ** isPrime(x)
        ** GCD(*nums) #Euclid's Algorithm for Greatest Common Divisior
        ** LCM(*nums) #Probably Another algorithm of Euclid, getting the Least Common Multiple
        ** FloatToIntegerDivison(num : float) converts float into integer division
        ** kCn(k,n) #k chose n, the binomial coefficient
"""

from .basic import product
from typing import Union,Tuple
from . import functions
from .trigonometric import arctan
from .constants import pi

def isEven(num : int) -> bool:
    """Returns True if a number can be divded by 2"""
    return num%2==0

def isOdd(num : int) -> bool:
    """Returns True if a number cannot be divded by 2"""
    return not isEven(num)

def isPrime(num : int) -> bool:
    """Returns True if a number can divide num in the \n
       ** range(2,int(1+num**(1/2))) **
       """
    if num == 1:
        return False

    for i in range(2,int(1+num**(1/2))):
        if(num%i==0):
            return False
    return True

def GCD_SIMPLE(num1 : int,num2 : int) -> int:
    """Find the greatest common multiple of 2 numbers"""
    divisor = num1 if num1 > num2 else num2
    dividor =  num1 if num1 != divisor else num2
    
    remainder = divisor%dividor
    times_in = divisor//dividor

    while remainder != 0:
        divisor = dividor
        dividor = remainder
        remainder = divisor%dividor
    
    return dividor

def GCD(*nums : int) -> int:
    """Uses Euclid's algorithm for finding the \n
       Greatest Common Multiple of a series of numbers, \n
       you can either pass arguments normally GCD(num1,num2,num3) \n
       or as a list using the * notation GCD(*myArray)
    """
    g1 = 0
    x = list(nums)
    while len(x) > 1:
        g1 = GCD_SIMPLE(x[0],x[1])
        x.pop(0)
        x.pop(0)
        x.insert(0,g1)
    return x[0]

def LCM(*args: Union[int, list]) -> int:
    """
        Given an array of integers or integers passed as arguments it can find the \n
        Least common multiple of those numbers, The smallest positive integer 'num_target' such that : \n
        (item%num_target==0 for item in args )
    """
    args = args[0] if len(args) == 1 else args

    # if the input is one number the outpout is the same number
    if (type(args) == int):
        return args

    state = {}
    nums = []

    # Create an Object keeping track of their state
    for item in args:
        state[item] = item

    def validState():
        """Checks if the object is in final format"""
        for item in state:
            if state[item] != 1:
                return False
        return True

    k = 1
    DIVISORS = []

    # Continue too loop through prime numbers while the state is not True
    while not validState():

        isAppended = False
        DIVISORS = []

        # Skip if not prime
        if not isPrime(k):
            k += 1
            continue

        for item in state:
            num = state[item]

            if num == 1:
                # Instance of object is in its final form
                DIVISORS.append(False)
                continue

            # when num is divisible by k we append k to nums[] and reduce the state
            if num % k == 0:
                state[item] = num / k  # State is reduced by k
                DIVISORS.append(True)
                if not isAppended:
                    nums.append(k)
                    isAppended = True
            else:
                DIVISORS.append(False)
        if not True in DIVISORS:
            k += 1
    return product(*nums)

def FloatToIntegerDivison(num : float) -> Tuple[int]:
    """Returns an a tuple of 2 integers, whose division
       produces the passed in floating point number
    """
    decimal_places = len(str(num).split(".")[1])
    expon = 10 ** decimal_places
    nom_denom = (expon * num,expon)
    g_c_d = GCD(nom_denom[0],nom_denom[1])
    while g_c_d > 1:
        nom_denom = (nom_denom[0] / g_c_d , nom_denom[1] / g_c_d)
        g_c_d = GCD(nom_denom[0],nom_denom[1])
    return (int(nom_denom[0]),int(nom_denom[1])) #Get rid of the .0 from the float division

def kCn(k : int,n : int) -> Union[int,float]:
    """Returns the binomial coefficient of 2 integer numbers"""
    # (7,3)
    return functions.factorial(k) / (functions.factorial(n) * functions.factorial(k-n) ) 


def complex_polar(z : complex):
    """Returns the the mangitude of the Vector
       of a complex number z and the angle it forms
       in the x-axis z : complex => [r,theta]
       from : r * (cos(x) + i*sin(x))
    """
    a = z.real;b= z.imag
    r = (a**2 + b**2)**(1/2)
    theta = arctan(b / a,iterations=50)
    if a < 0:
        theta += pi
    return [r,theta] # r * e**(i*theta)


if __name__ == "__main__": 
    pass


__all__ = [
    "basic",
    "constants",
    "functions",
    "linear",
    "powers",
    "random",
    "polynomials",
    "num_theory",
    "trigonometric",
]

from . import basic
from . import constants
from . import functions
from . import linear
from . import powers
from . import random
from . import polynomials
from . import num_theory
from . import trigonometric


"""
    Submodule for calculations involing trigonometric functions,\n
    It calculates all the functions from scratch,\n
    using some form of 'infinite' summation (Taylor Expansion most of the time)\n

    It can also handle complex numbers in non inverse functions (arc functions):
        ** for more information on how they are computed see https://www.youtube.com/watch?v=CjQTWtW_x9o

    
    Here is a detailed list of what is here:
        - Conversions:
            Degrees TO Rad
            Rad TO Degrees
        - Trigonometric (Real or Complex):
            ** sin(x) 
            ** cos(x) 
            ** tan(x) 
            ** cot(x) 
            ** sec(x) 
            ** csc(x) 
        - Inverse Trigonometric (Real Only for accurate results):
            ** arcsin(x)
            ** arccos(x)
            ** arctan(x)
            ** arccot(x)
            ** arcsec(x) 
            ** arccsc(x)  
        -- Hyperbolic Trigonometric (Real and Imaginary):
            ** sinh(x)
            ** cosh(x)
            ** tanh(x)
            ** coth(x)
            ** csch(x)
            ** sech(x)
        - Inverse Hyperbolic (Real only for accurate results):
            ** arsinh(x)
            ** arcosh(x)
            ** artanh(x)
            ** arcoth(x)
            ** arcsch(x)
            ** arsech(x)
"""

# NOTE : MORE THAN 13 ITERATIONS GIVES NOTHING AND IT REDUCES PERFORMANCE

from . import functions
from .constants import pi,e
from . import powers
from typing import Union

rad360 = 6.283185307179586 #360 degrees in radians

# Conversions
def toRad(degrees):
    """Degrees to rad"""
    return (pi/180)*degrees

def toDegrees(rad):
    """Rad to degrees"""
    return (180 / pi) * rad

#For handling angles greater than 360 deg and rad360 radians (this allows to reduce iterations to 13 for trig functions)
def reduceAngle(trig_function):
    """Given a value that is more than 360 it finds it's coressponding angle in range (0,360)"""
    def validator(*args,**kwargs):
        if not type(args[0]) == complex:
            rads = not kwargs.get('degrees') if kwargs.get('degrees') is not None else True
            iterations = kwargs.get('iterations')
            angle = args[0]
            if rads:
                if abs(angle) > rad360:
                    sub = (angle // rad360) *rad360
                    angle =  angle-sub
            if abs(angle) > 360:
                sub = (angle // 360) * 360
                angle = angle - sub
            return trig_function(angle,kwargs.get('degrees'),iterations if iterations is not None else 50)
        return trig_function(*args,**kwargs)

    return validator

# Trigonometric
@reduceAngle
def sin(x : Union[float,complex],degrees=False,iterations : int = 13) -> Union[float,complex]:
    """ Trigonometric function : Sine\n
        Domain : All Real\n
        You can specify whether you want to use degrees or radians by passing in degrees=bool (default is False)\n
        You can also specify how many times you want to iterate by passing in iterations=int\n
        it uses Taylor expansions and trigonometric identities for caluclations\n
        This asian explains how it treats complex numbers : https://www.youtube.com/watch?v=CjQTWtW_x9o
        """
    if type(x) == complex:
        com = complex(0,1) * x
        return (powers.power(e,com) - powers.power(e,-com)) / 2 * complex(0,1)
    if degrees:
        x = toRad(x)
    #Taylor series for sin x
    total = 0
    for i in range(iterations):
        alternating = (-1)**i
        denominator = functions.factorial(2*i+1)
        alternating_denominator = alternating / denominator
        input_adjust = x**(2*i+1)
        total += alternating_denominator * input_adjust
    return total

@reduceAngle
def cos(x: Union[float,complex],degrees=False,iterations : int = 13) -> Union[float,complex]:
    """ Trigonometric function : Cosine\n
        Domain : All Real\n
        You can specify whether you want to use degrees or radians by passing in degrees=bool (default is False)\n
        You can also specify how many times you want to iterate by passing in iterations=int\n
        it uses Taylor expansions and trigonometric identities for caluclations
        """
    if type(x) == complex:
        com = complex(0,1) * x
        return (powers.power(e,com) - powers.power(e,-com)) / 2

    if degrees:
        x = toRad(x)
    reduced_pi = pi / 2
    return sin(reduced_pi-x,iterations=iterations)

@reduceAngle
def tan(x: Union[float,complex],degrees=False,iterations : int = 13) -> Union[float,complex]:
    """ Trigonometric function : Tangent\n
        Domain : All numbers whose cos(x) is not 0\n
        You can specify whether you want to use degrees or radians by passing in degrees=bool (default is False)\n
        You can also specify how many times you want to iterate by passing in iterations=int\n
        it uses Taylor expansions and trigonometric identities for caluclations
        """
    if degrees:
        x = toRad(x)
    return sin(x,iterations=iterations) / cos(x,iterations=iterations)

@reduceAngle
def cot(x: Union[float,complex],degrees=False,iterations : int = 13) -> Union[float,complex]:
    """ Trigonometric function : Cotangent\n
        Domain : All numbers whose sin(x) is not 0\n
        You can specify whether you want to use degrees or radians by passing in degrees=bool (default is False)\n
        You can also specify how many times you want to iterate by passing in iterations=int\n
        it uses Taylor expansions and trigonometric identities for caluclations
        """
    if degrees:
        x = toRad(x)
    return 1 / tan(x,iterations=iterations)

@reduceAngle
def sec(x: Union[float,complex],degrees=False,iterations : int = 13) -> Union[float,complex]:
    """ Trigonometric function : Secant\n
        Domain : All numbers whose cos(x) is not 0\n
        You can specify whether you want to use degrees or radians by passing in degrees=bool (default is False)\n
        You can also specify how many times you want to iterate by passing in iterations=int\n
        it uses Taylor expansions and trigonometric identities for caluclations
        """
    if degrees:
        x = toRad(x)
    return 1 / cos(x,iterations=iterations)

@reduceAngle
def csc(x: Union[float,complex],degrees=False,iterations : int = 13) -> Union[float,complex]:
    """ Trigonometric function : Cosecant\n
        Domain : All numbers whose sin(x) is not 0\n
        You can specify whether you want to use degrees or radians\n
        You can also specify how many times you want to iterate\n
        it uses Taylor expansions and trigonometric identities for caluclations
        """
    if degrees:
        x = toRad(x)
    return 1 / sin(x,iterations=iterations)

# Inverse Trigonometric

def arcsin(x : float,iterations : int = 100,degrees : bool = False) -> float:
    """
        Inverse Trigonometric function : Arcsin\n
        equivalant to the following expression : \n
        ** x = sin(y) => arcsin(x) = y **\n
        if deegrees is set to True it will give the result in degrees\n
        Domain (-1 <= x <= 1)
    """
    if not (-1 <= x <= 1):
        raise ValueError("Math domain error not in [-1,1]")
    total = 0
    for n in range(iterations):
        nominator_0 = functions.factorial(2*n)
        nominator_1 = powers.power(x,2*n+1)
        denominator_0 = powers.power(powers.power(2,n)*functions.factorial(n),2)
        denominator_1 = 2*n+1
        div_0 = nominator_0 / denominator_0
        div_1 = nominator_1 / denominator_1
        total += div_1 * div_0
    if degrees:
        total = toDegrees(total)
    return total

def arccos(x : float,iterations : int = 100,degrees : bool = False) -> float:
    """
        Inverse Trigonometric function : Arccos\n
        equivalant to the following expression : \n
        ** x = cos(y) => arccos(x) = y **\n
            if deegrees is set to True it will give the result in degrees\n
        Domain (-1 <= x <= 1)
    """
    if not (-1 <= x <= 1):
        raise ValueError("Math domain error not in [-1,1]")
    result = (pi / 2) - arcsin(x,iterations=iterations)
    if degrees:
        result = toDegrees(result)
    return result

def arctan(x : float,iterations : int = 100,degrees : bool = False) -> float:
    """
        Inverse Trigonometric function : Arctangent\n
        equivalant to the following expression : \n
        ** x = tan(y) => arctan(x) = y **\n
        if deegrees is set to True it will give the result in degrees\n
        Domain : All real numbers
    """
    forumlae = arccos(  (1-x**2) / (1 + x**2)    )
    total = 0.5 * forumlae
    if degrees:
        total = toDegrees(total)
    return total

def arccot(x : float,iterations : int = 100,degrees : bool = False):
    """
        Inverse Trigonometric function : Arccotangent\n
        equivalant to the following expression : \n
        ** x = cot(y) => arccot(x) = y **\n
        if deegrees is set to True it will give the result in degrees\n
        Domain : All real numbers
    """
    result = (pi / 2) - arctan(x,iterations=iterations)
    if degrees:
        result = toDegrees(result)
    return result

def arcsec(x : float,iterations : int = 100,degrees : bool = False):
    """
        Inverse Trigonometric function : Arcsecant\n
        equivalant to the following expression : \n
        ** x = sec(y) => arcsec(x) = y **\n
        if deegrees is set to True it will give the result in degrees\n
        Domain : (x <= -1 or x >= 1)
    """
    if (x <= -1) or (x >= 1):
        res =  arccos(1/x,iterations=iterations)
        if degrees:
            res = toDegrees(res)
        return res
    else:
        raise ValueError("Math domain error not in (x <= -1 or x >= 1)")


def arccsc(x : float,iterations : int = 100,degrees : bool = False):
    """
        Inverse Trigonometric function : Arccosecant\n
        equivalant to the following expression : \n
        ** x = csc(y) => arccsc(x) = y **\n
        if deegrees is set to True it will give the result in degrees\n
        Domain : (x <= -1 or x >= 1)
    """
    if (x <= -1) or (x >= 1):
        result = ( pi / 2 ) - arcsec(x)
        if degrees:
            result = toDegrees(result)
        return result
    else:
        raise ValueError("Math domain error not in (x <= -1 or x >= 1)")


# Hyperbolic Trigonometric

def sinh(x : float,useTaylor : bool = False,iterations: int = 100) -> float:
    """
        Hyperbolic Sine,\n
        domain : 'All real numbers',\n
        if useTayolor is true it will use the taylor expansion,\n
    """
    if useTaylor:
        return sum([
            powers.power(x,2*n+1) / functions.factorial(2*n+1) for n in range(iterations)
        ])
    return (powers.power(e,x) - powers.power(e,-x)) / 2

def cosh(x : float,useTaylor : bool = False,iterations = 100) -> float:
    """
        Hyperbolic Cosine,\n
        domain : 'All real numbers',\n
    """
    if not useTaylor:
        return (powers.power(e,x) + powers.power(e,-x)) / 2
    return sum([
        powers.power(x,2*n) / functions.factorial(2*n) for n in range(iterations)
    ])

def tanh(x : float) -> float:
    """
        Hyperbolic Tangent,\n
        domain : 'All values whose cosh(x) != 0',\n
    """
    return sinh(x) / cosh(x)

def coth(x : float) -> float:
    """
        Hyperbolic Tangent,\n
        domain : 'All values whose tanh(x) != 0',\n
    """
    return 1 / tanh(x)

def sech(x : float) -> float:
    """
        Hyperbolic Tangent,\n
        domain : 'All values whose cosh(x) != 0',\n
    """
    return 1 / cosh(x)

def csch(x : float) -> float:
    """
        Hyperbolic Tangent,\n
        domain : 'All values whose sinh(x) != 0',\n
    """
    return 1 / sinh(x)

# Inverse Hyperbolic

def arsinh(x : float) -> float:
    """
        Inverse Hyperbolic trigonometric function\n
        Domain : All Real
    """
    return functions.ln(x+powers.sqrt(powers.power(x,2)+1))

def arcosh(x : float) -> float:
    """
        Inverse Hyperbolic trigonometric function
        Domain : [1, +Infinity)
    """
    if x < 1:
        raise ValueError("Math domain error [1,+Infinity]")
    return functions.ln(x+powers.sqrt(powers.power(x,2)-1))

def artanh(x : float) -> float:
    """
        Inverse trigonometric function
        Domain (-1 < x < 1)
    """
    if not -1 < x < 1:
        raise ValueError("Math domain error not in  (-1,+1)")
    return 0.5*functions.ln((1+x)/(1-x))

def arcoth(x : float) -> float:
    """
        Inverse Hyperbolic trigonometric function
        (-Infinity, −1) and (1, +Infinity)
    """
    if x == 1 or x == -1:
        raise ValueError("Math domain error not in  (-Infinity, −1) and (1, +Infinity)")
    return 0.5*functions.ln((x+1)/(x-1))

def arsech(x : float) -> float:
    """
        Inverse Hyperbolic trigonometric function
        Domain : (0, 1]
    """
    if x >= 0 or  x > 1:
        raise ValueError("Math domain error not in (0, 1]")
    return functions.ln( (1+powers.sqrt(1-powers.power(x,2)) / x ))

def arcsch(x : float) -> float:
    """
        Inverse Hyperbolic trigonometric function
        Domain : All Real Numbers except 0
    """
    if x == 0:
        raise ValueError("Math domain error not in (All Real Numbers except 0)")
    return functions.ln( (1/x) + powers.sqrt((1/powers.power(x,2)) + 1) )

def complexRoot(n):
    """For negative Nth roots"""
    return cos(pi/n) + complex(0,1) * sin(pi / n)

if __name__ == "__main__":
    pass    


"""
    Contains a series of functions and custom classes,\n
    concerning the base of linear algebra,
    Matrices and Vectors

    Here are the operations:
    -- Vectors : 
        ** Vector Dot 
        ** Vector Cross (3D ONLY)
        ** Vector Addition and Subtraction
        ** Vector-Scalar Operations
    -- Matrices:
        ** Matrix Multiplication
        ** Matrix Addition and subtraction
        ** Matrix-Scalar Operations
        ** Matrix-Vector Operations
        ** Trace
        ** Identity Matrix Generator
        ** Determinant
        ** REF (Reduced Echelon Form)
        ** Inverse Matrix
            ** Cofactors
            ** adjugate (transpose)
            ** Minors
        *** Various combinations of the above
"""

from .basic import isNumber,isInteger,isComplex,Number
from . import polynomials as pl
from .trigonometric import arccos
from .powers import sqrt
from .num_theory import complex_polar
import re
from typing import Union,Any,Dict,Tuple

WHITESPACE = ' '


epsilon = pl.x #epsilon = lamda 
POLY = type(epsilon) #<class 'pythematics.polynomials.Polynomial'>


class Vector:
    def __init__(self,array):
        for item in array:
            if not isNumber(item) and type(item) != complex and type(item) != POLY:
                raise ValueError("Vector arguments must be a list of {} , {} , {} or {} not {}".format(int,float,complex,POLY,type(item)))
        self.matrix = array
        self.rows = len(self.matrix)
        self.collumns = 1

    def __str__(self):
        print("")
        i = 1
        for item in self.matrix:
            if 'deg' in str(item): #Check if polynomial
                item = re.sub(r"\s+",'',str(item).split(":")[-1])
                item = f'({item})'
            print(f'R{i}| {item:>3}')
            i+=1
        s2 = "\n{} x {} Vector array\n".format(self.rows,self.collumns)
        return s2

    def getMatrix(self):
        return self.matrix

    def getSize(self):
        return self.rows

    def __getitem__(self,index):
        return self.matrix[index]

    def __add__(self,value):
        empty = []
        if type(value) == Vector:
            if value.getSize() != self.getSize():
                raise ValueError("Cannot multiply non equal-size collumns ({} with {})".format(value.getSize(),self.getSize()))
            for i in range(self.getSize()):
                empty.append(value.getMatrix()[i] + self.getMatrix()[i])
            return Vector(empty)
        return NotImplemented

    def __sub__(self,value):
        empty = []
        if type(value) == type(self):
            if value.getSize() != self.getSize():
                raise ValueError("Cannot multiply non equal-size collumns ({} with {})".format(value.getSize(),self.getSize()))
            for i in range(self.getSize()):
                empty.append(value.getMatrix()[i] - self.getMatrix()[i])
            return Vector(empty)
        else:
            return NotImplemented

    def __len__(self):
        return self.rows

    def __mul__(self,value):
        """Vector Multiplication by scalar
            if other value is Vector,
            the dot product is returned
        """
        empty = []
        if isNumber(value) or type(value) == complex or type(value) == type(epsilon): #Scalar
            for item in self.matrix:
                empty.append(value*item)
            return Vector(empty)
        elif type(value) == type(self):
            if value.getSize() != self.getSize():
                raise ValueError("Cannot multiply non equal-size collumns ({} with {})".format(value.getSize(),self.getSize()))
            for num in range(self.getSize()):
                empty.append(value.getMatrix()[num] * self.getMatrix()[num])
            return sum(empty)
        elif type(value) == Matrix:
            vector_to_matrix = Matrix([[item] for item in self.getMatrix()])
            return vector_to_matrix * value

        return NotImplemented #Redefine with __rmul__

    def __neg__(self):
        return (-1) * self

    def __rmul__(self,scalar : Union[int,float]):
        if type(scalar) in (int,float,complex,POLY):
            return self.__mul__(scalar)
        raise TypeError("Cannot perform '*' operation on Vector with {}")

    def __round__(self,ndigits : int = 1):
        __tmp__ : list = []
        for item in self.getMatrix(): 
            if type(item) is complex:
                rounded = complex(round(item.real,ndigits),round(item.imag,ndigits))
            else:
                rounded = round(item,ndigits)
            __tmp__.append(rounded)
        return Vector(__tmp__)

    def dot(self,Vector) -> Union[float,int]:
        return self.__mul__(Vector) 

    def cross(self,Vector : 'Vector') -> 'Vector':
        return cross(self,Vector)

    def magnitude(self):
        return magnitude(self)
    
    def AngleVector(self,vector1 : "Vector",degrees : bool = False) -> float:
        return AngleBetweenVectors(self,vector1,degrees)

    # def cross(self,Vector)

class Matrix:
    """
    The as you known it from math 'Matrix'\n
    It includes custom operations for the following:
        ** Multiplication
        ** Addition
        ** Subtraction

    These are also supported not as methods but as seperate functions:
        ** determinant
        ** inverse
        ** Transpose (adjugate)
        ** Matrix of co-factors (Alternating-Chess pattern sign)
        ** Matrix of  Minors (for each element hide the current row and collumn and find the determinant of the following) 

    You must pass an array of arrays,\n
    Inside the base array the nested lists are considered as the rows,\n
    and the collumns are determined vertically\n
    Matrices shall be passed in the following format :

    [   
            #Col1  #Col2  #Col3
     #Row 1 [num1 , num2 , num3 ...  numP],
     #Row 2 [num4,  num5 , num6 ...  numP],
                ..........      ...
     #Row n [numK,numL,numO     ...  numP]
    ]

        Input :
        A = Matrix([
            [1,2,3],
            [4,5,6],
            [7,8,9]
        ])

        Output :
                C1      C2      C3

        R1 |    1        2       3
        R2 |    4        5       6
        R3 |    7        8       9

    for more specific methods you can use the dir() function

    """
    def __init__(self,matrix):
        """Takes an array of arrays of numbers
            The arrays inside the array as seen as the rows
        """
        if type(matrix) != list:
            raise ValueError("Matrix must be an array of arrays.")
        
        self.ROW_LENGTHS = []

        for row in matrix:
            if type(row) == list:
                self.ROW_LENGTHS.append(len(row))
                for num in row:
                    if not isNumber(num) and not type(num) == type(epsilon) and not type(num) == complex:
                        raise TypeError("Row : {} , does not contain an argument or {} or {} but instead {}!".format(matrix.index(row),type(1),type(1.0),type(num)))
            else:
                raise ValueError("Every argument inside the base array which is considered as a row should be of type {} (Your array had at least one element that was not of type {})".format(list,list))
        if len(self.ROW_LENGTHS) != self.ROW_LENGTHS.count(self.ROW_LENGTHS[0]):
            raise ValueError("All rows of a matrix shall only be of same size. not {}".format(self.ROW_LENGTHS))

        self.matrix = matrix
        self.rows = len(self.matrix)
        self.collumns = self.ROW_LENGTHS[0]
        self.isSquare = self.rows == self.collumns

        self.cols = []
        for j in range(self.ROW_LENGTHS[0]):
            self.cols.append([])

        for row in self.matrix:
            i = 0
            for value in row:
                self.cols[i].append(value)
                i+=1

    def rawMatrix(self):
        """Returns the raw array passed in (self.matrix)"""
        return self.matrix

    def colls(self,index : int = 0) -> list:
        """Returns a collumn when an index is specified (default is 0)"""
        return self.cols[index]
    
    def is_square(self) -> bool:
        return self.isSquare

    def collsAll(self) -> list:
        """Returns an array of all the collumns"""
        return self.cols

    def row(self,index : int = 0):
        """Returns a specific row given an index (default is 0)"""
        return self.matrix[index]

    def index(self,row,collumn):
        """Returns the position given the corresponding row and collumn"""
        return self.matrix[row][collumn]

    def __len__(self):
        """Returns a tuple containng number of rows and collumns (rows,collumns)"""
        return (self.rows,self.collumns) # (number of rows,number of collumns)

    def __eq__(self,value):
        if type(value) in (type(self),Vector):
            array_item = value.rawMatrix() if type(value) == type(self) else value.getMatrix()
            return self.rawMatrix() == array_item
        return NotImplemented

    def __str__(self):
        print("")
        x = [item[:] for item in self.matrix]
        k = 0
        for item in x:
            j = 0
            if len(item) > 8:
                x[k] = item[1:9]
                x[k].append("...")
                x[k].append(self.cols[-1][k])
                j+=1
        
            k+=1
        k = 0        
        y = []    
        for iteration in range(self.collumns):
            if iteration >=8:
                y.append("...")
                y.append(f'C{self.collumns}')
                break
            y.append(f"C{iteration+1}")
        x.insert(0,y)
        j = 1
        
        for item in x:
            if j > 9:
                print("\n   .........")
                CACHE = []
                for item in x[-1]:
                    if Number(item):
                        if type(item) == complex:
                            y = complex_polar(item)
                            NEW_ARRAY.append(f"({round(y[0],2)},{round(y[1],2)})")
                            continue
                        CACHE.append('{: <10}'.format(round(item,4)))
                        continue
                    CACHE.append('{: <10}'.format(item))
                print(f' R{len(x)-1}|',*CACHE)
                break
            item[0] = f'\t{item[0]}'
            if j==1:
                cols_t = " ".join(["{: <10}" for _ in range(len(item))])
                cols_t = cols_t.format(*item)
                print(' CI | {}'.format(cols_t))
                j+=1
                continue
            NEW_ARRAY = []
            for val in item:
                if "..." in str(val):
                    NEW_ARRAY.append(val)
                    continue
                if not 'deg' in str(val):
                    val = complex(val)

                    if val.imag != 0:                    
                        com_val = complex(val)
                        y = complex_polar(com_val)
                        NEW_ARRAY.append(f"({round(y[0],2)},{round(y[1],2)})")
                        continue

                    test = val.real

                    if test != int(test):
                        float_rounded = float(test)
                        if len(str(float_rounded)) >= 10:
                            value = round(float_rounded,2)
                            if len(str(value)) >=10:
                                value = f'{float(val):.2e}'
                        else:
                            value = f'{float_rounded:>4}'
                    else:
                        if len(str(val)) >= 10:
                            value = f'{int(test):.2e}'
                        else:
                            value = f'{int(test):>3}'
                    NEW_ARRAY.append(value)
                    continue

                else:
                    ws_pol = str(val).split(":")[-1]
                    no_ws_pol = re.sub(r"\s+",r"",ws_pol)
                    finale = f'({no_ws_pol})'
                    if len(finale) <= 7:
                        NEW_ARRAY.append(f'({no_ws_pol})')
                    elif 7 < len(finale) <= 9 :
                        NEW_ARRAY.append(f'{no_ws_pol}')
                    else:
                        NEW_ARRAY.append(f'({no_ws_pol})')

            
            cols_t = " ".join(["{: <10}" for _ in range(len(NEW_ARRAY))])
            print(f' R{j-1} |',cols_t.format(*NEW_ARRAY))
            j+=1
        return f'\n{self.rows} x {self.collumns} Matrix\n'

    def __round__(self,ndigits : int = 1) -> "Matrix":
        __tmp__ : list = [[] for item in self.rawMatrix()]
        i = 0
        for row in self.rawMatrix():
            for item in row: 
                if type(item) is complex:
                    rounded = complex(round(item.real),round(item.imag))
                else:
                    rounded = round(item,ndigits)

                __tmp__[i].append(rounded)
            i+=1
        return Matrix(__tmp__)


    def __rmul__(self,scalar):
        """Matrix multiplication by scalar"""
        if type(scalar) in (int,float,complex,type(epsilon)):
            new_matrix = [[] for i in range(self.rows)] #Add the rows
            i = 0
            for row in self.matrix:
                for constant in row:
                    new_matrix[i].append(constant * scalar)
                i+=1
            return Matrix(new_matrix)
        elif type(scalar) == Vector:
            return self.__mul__(adjugate(Matrix(scalar.getMatrix())))
        raise TypeError("You may only multiply a {} object with either a {} or a {}".format(type(self),int,float))

    def __neg__(self):
        return (-1) * self
    
    def __add__(self,Matrx):
        if type(Matrx) != type(self):
            raise ValueError("A Matrix may only be added with another Matrix not {}!".format(type(Matrx)))
        if self.__len__() != Matrx.__len__():
            raise ValueError("Rows and Collumns must be equal! {} != {}".format(self.__len__(),Matrx.__len__()))
        new_matrix = [[] for row in range(self.rows)]
        i = 0
        for row in self.matrix:
            k = 0 
            for num in row:
                new_matrix[i].append(num+Matrx.rawMatrix()[i][k])
                k +=1
            i+=1
        return Matrix(new_matrix)

    def __sub__(self,Matrx):
        if type(Matrx) != type(self):
            raise ValueError("A Matrix may only be added with another Matrix not {}!".format(type(Matrx)))
        if self.__len__() != Matrx.__len__():
            raise ValueError("Rows and Collumns must be equal! {} != {}".format(self.__len__(),Matrx.__len__()))
        new_matrix = [[] for row in range(self.rows)]
        i = 0
        for row in self.matrix:
            k = 0 
            for num in row:
                new_matrix[i].append(num-Matrx.rawMatrix()[i][k])
                k +=1
            i+=1
        return Matrix(new_matrix)

    def __mul__(self,value):
        if type(value) in (int,float,complex,type(epsilon)):
            return self.__rmul__(value)
        if type(value) == Vector:
            vector_to_matrix = Matrix([[item] for item in value.getMatrix()])
            return self * vector_to_matrix
        row_0 = self.__len__()
        col_0 = value.__len__()
        if row_0[1] != col_0[0]: 
            raise ValueError(f"\nCannot multiply a {row_0[0]} x {row_0[1]} with a {col_0[0]} x {col_0[1]} Matrix,\nMatrix 1 must have the same number of rows as the number of collumns in Matrix 2 \n({row_0[1]} != {col_0[0]})")
        new_matrix = [[] for i in range(row_0[0])]
        COLS_M2 = value.collsAll()
        j = 0
        for row in self.matrix:
            for collumn in COLS_M2:
                iterations = 0
                total = 0
                for scalar in collumn:
                    total += scalar*row[iterations]
                    iterations+=1
                new_matrix[j].append(total)
            j+=1
        return Matrix(new_matrix)

    def __getitem__(self,index):
        return self.rawMatrix()[index]

    def appendCollumn(self,collumn : list) -> None:
        assert len(collumn) ==  len(self.colls(0)), "New Collumn must be of the same size as all the other collumns"
        new_matrix = [item[:] for item in self.rawMatrix().copy()]
        for i in range(len(collumn)):
            new_matrix[i].append(collumn[i])
        return Matrix(new_matrix)

    def transpose(self):
        """Evaluates the function adjugate(self)"""
        return adjugate(self)
    
    def determinant(self):
        """Evaluates the function determinant(self)"""
        return determinant(self)

    def minors(self):
        """evaluates the function MatrixOfMinors(self)"""
        return MatrixOfMinors(self)

    def cofactors(self):
        """evaluates the function MatrixOfCofactors(self)"""
        return MatrixOfCofactors(self)

    def inverse(self):
        """evaluates inverse(self)"""
        return inverse(self)

    def trace(self):
        """evaluates Trace(self)"""
        return Trace(self)

    def swap(self,Row1 : int,Row2 : int):
        """Swaps 2 rows given their index"""
        val_0=[item for item in self.rawMatrix()[Row1]]
        val_1=[item for item in self.rawMatrix()[Row2]]
        self.rawMatrix()[Row1] = val_1
        self.rawMatrix()[Row2] = val_0

    def solve(self,Output : Vector,unknowns : Union[tuple,list],useRef=False) -> dict:
        """Solves the system of linear equations represented by the current Matrix\n
           An 'Output' Vector should be provided in order for the augmented Matrix to complete,\n
           and also the Name of the unknowns should be provided in order to receive the solutions in order\n
           You can also specify useRef=bool on wheter you want to use Row reduction, (if it is et to false
           it uses Cramer's rule)

           EXAMPLE : 
            A = Matrix([
                [1,2],
                [3,4]
            ])

            unknowns = ('x','y')
            output = Vector([5,11])
            solution = A.solve(output,unknowns)
            print(solution)

           OUTPUT :
            {'x': 1.0, 'y': 2.0}
        """
        if not useRef:
            return SolveCramer(self,unknowns,Output)
        return solveREF(self,unknowns,Output)

    def ref(self):
        """Returns the Redcued Echelon Form of the Matrix"""
        return ref(self)

    def CharacteristicPolynomial(self):
        """Returns a Polynomial whose Roots are
           the eigenvalues of that Matrix
        """
        return CharacteristicPolynomial(self)

    def eigen_values(self,iterations : int = 50):
        """Find the eigenvalues of the Matrix
           given that it is square\n
        """
        return eigenValues(self,iterations)

    def rank(self):
        """Returns the number of linearly independent rows"""
        return rank(self)

    def forEach(self,function : callable,applyChanges : bool = False) -> Union['Matrix',None]:
        """For each element of the matrix it performs a given function on that element\n
           and it either returns a new transform matrix if applyChanges = False,\n
           otherwise it returns Nothing and applies the changes to the given Matrix\n
        """
        BufferArray = [[] for item in self.matrix]
        i = 0
        for row in self.matrix:
            for num in row:
                BufferArray[i].append(function(num))
            i+=1
        if not applyChanges:
            return Matrix(BufferArray)
        self.matrix = BufferArray

def removeCollumn(matrix : Matrix,index : int) -> Matrix:
    """Returns a reduced collumn version of a Matrix"""
    raw_matrix = [item[:] for item in matrix.rawMatrix()]
    for row in raw_matrix:
        row.pop(index)
    return Matrix(raw_matrix)

def determinant(matrix : Matrix) -> float:
    dimensions = matrix.__len__()
    if not matrix.is_square():
        raise ValueError("Cannot compute determinant of non square matrix : {}".format(dimensions))
    if dimensions[0] == 2:
        return matrix.rawMatrix()[0][0] * matrix.rawMatrix()[-1][-1] - matrix.rawMatrix()[0][-1]* matrix.rawMatrix()[-1][0]
    raw_matrix = matrix.rawMatrix()
    tmp = [item[:] for item in raw_matrix]
    tmp.pop(0)
    i = 0 
    STORAGE = []
    for i in range(matrix.__len__()[0]): #Loop throw the first row
        y = removeCollumn(Matrix(tmp),i)
        multiplier = raw_matrix[0][i] if (i+1)%2!=0 else -raw_matrix[0][i]
        STORAGE.append(multiplier * determinant(y))
        i+=1
    return sum(STORAGE)

def MatrixOfCofactors(matrix : Matrix) -> float:
    """
        Given any NxM Matrix \n : 
        it reutrns a new Matrix,
        that follows the chessboard pattern
    """

    if matrix.__len__()[0] == 2:
        raise ValueError("Matrix must be more than 2 dimensional")
    array = [item[:] for item in matrix.rawMatrix()]
    new_array = [[] for item in matrix.rawMatrix()]
    i = 0
    positive = True
    positive_col = True
    for row in array:
        j = 0
        for number in row:
            if positive:
                new_array[i].append(number)
            else:
                new_array[i].append(-number)

            if j+1 != len(row):
                positive = not positive
            
            else:
                positive_col = not positive_col
                positive = positive_col
            j+=1
        i+=1
    return Matrix(new_array)

def adjugate(matrix : Matrix) -> float:
    """It transposes a given Matrix,"""
    array = [item[:] for item in matrix.rawMatrix()]
    arrays = [[] for item in matrix.collsAll()]
    for row in array:
        i = 0
        for num in row:
            arrays[i].append(num)
            i+=1
    return Matrix(arrays)

def MatrixOfMinors(matrix : Matrix) -> Matrix:
    """
        Given an square Matrix that is not 2x2 it returns a new Matrix,\n
        The new Matrix is generated by the determinants generated by the following:\n
            ** For each item in the Matrix :
                ** 'Hide' the current collumn and row
                ** Now compute the determinant of the remaining Matrix
    """
    matrix_len = matrix.__len__()
    if not matrix.is_square():
        raise ValueError("Cannot perfrom Matrix of minors on non-square matrix : {}".format(matrix_len))
    matrix_array = [row[:] for row in matrix.rawMatrix()]
    j=0
    DETERMINANTS = [[] for row in matrix.rawMatrix()]
    for row in matrix_array:
        i = 0 
        reduced = [item[:] for item in matrix_array]
        reduced.pop(j)
        for num in row:
            x = removeCollumn(Matrix(reduced),i)
            DETERMINANTS[j].append(determinant(x))
            i+=1
        j+=1
    return Matrix(DETERMINANTS)

def inverse(matrix : Matrix) -> Matrix:
    """
    Returns the inverse of a Matrix, if it is invertible (non-zero determinant)

        \n=> Find 'Matrix of Minors'; #New Matrix with the determinants of each item of the array
        \n=> Find Matrix of co-factors of the previous Matrix; #Alternating chessboard sign
        \n=> Transpose (adjugate) that Matrix
        \n=> Multiply by 1 / determinant
    """
    assert matrix.is_square() , "Cannot Invert non square matrix : {}".format(matrix.__len__())
    if matrix.__len__()[0] == 2:
        raw = matrix.rawMatrix()
        return (1 / determinant(matrix)) * Matrix(
            [[raw[-1][-1],-raw[0][-1]],
            [-raw[-1][0],raw[0][0]]
        ])
    try:
        inverse_determinant = 1 /  determinant(matrix)
    except:
        raise ZeroDivisionError("Matrix is not invertible due to it's determinant being zero")
    return inverse_determinant * adjugate(MatrixOfCofactors(MatrixOfMinors(matrix)))

def cross(vector_1 : Vector,vector_2 : Vector) -> Vector:
    if (type(vector_1),type(vector_2)).count(Vector) != 2:
        raise TypeError("Both arguments must be Vectors not {} and {}".format(type(vector_1),type(vector_2)))
    if (len(vector_1.getMatrix()),len(vector_2.getMatrix())).count(3) != 2:
        raise ValueError("Cannot perform cross product on non 3-dimensional Vectors : ({},{})".format(len(vector_1.getMatrix()),len(vector_2.getMatrix())))
    A = [vector_1.getMatrix(),vector_2.getMatrix()]
    DETERMINANTS = []
    for i in range(3):
        if (i+1)%2==0:
            DETERMINANTS.append(-determinant(removeCollumn(Matrix(A),i)))
            continue
        DETERMINANTS.append(determinant(removeCollumn(Matrix(A),i)))
    return Vector(DETERMINANTS)

def IdenityMatrix(dimensions : int) -> Matrix:
    if dimensions <= 1:
        raise ValueError("Dimensions must be at least 2 (not {}).".format(dimensions))
    matrix = []
    for i in range(dimensions):
        row = []
        for k in range(dimensions):
            if k == i:
                row.append(1)
                continue
            row.append(0)
        matrix .append(row)
    return Matrix(matrix)

def Trace(matrix : Matrix) -> Union[int,float]:
    """Returns the sum of the diagnals of a matrix"""
    if type(matrix) != Matrix:
        raise TypeError("Cannot only perform 'Trace' operation on {} (not {})".format(Matrix,type(matrix)))
    if not matrix.is_square():
        raise ValueError("Cannot only perform 'Trace' operation square matrices (not {})".format(matrix.__len__()))
    raw_matrix = matrix.rawMatrix()
    diagnals = []
    i = 0 #Track of row_matrix.index(row)
    for row in raw_matrix: 
        j = 0
        for num in row:
            if j==i:
                diagnals.append(num)
                break
            j+=1
        i+=1
    return sum(diagnals)

def isSwappable(row : Union[list,int]) -> bool:
    for item in row:
        if not isNumber(item):
            return False
        return True

def swap(matrix : Matrix,row1 : Union[list,int],row2 : Union[list,int]):
    """Swapws rows given a list (containg  the lements of the row) or the indexes of the rows (RETURNS A COPY OF THE NEW MATRIX)\n
       it DOESN'T handle duplicates\n
       if you want no matter what to switch rows use the self.swap() function\n
    """
    assert type(row1) in (int,list) and type(row2) in (int,list), "Row must either be a list or an index"
    i = 0
    for row in [row1,row2]:
        if type(row) == list:
            assert isSwappable(row1),"Instances of classes that are not {} or {} were found".format(int,float) #Check if it contain numbers
        else:
            if i == 0:
                row1 = matrix[row] #Set it equal to the position
            else:
                row2 = matrix[row]
        i+=1
    rows = [item[:] for item in matrix.rawMatrix()]
    is_duplicate_1 = True if rows.count(row1) > 1 else False
    is_duplicate_2 = True if rows.count(row2) > 1 else False
    if is_duplicate_1 or is_duplicate_2:
        if is_duplicate_1 and is_duplicate_2:
            duplicate = "Row 1 and Row 2 :  {},{}\n were".format(row1,row2)
        else:
            duplicate = "Row 1 : {} was".format(row1) if is_duplicate_1 else "Row 2 : {} was".format(row2)

        raise ValueError(f'{duplicate} found more than once in the Matrix.')
    #Get each index
    index_1 = rows.index(row1)
    index_2 = rows.index(row2)
    rows[index_1] = row2
    rows[index_2] = row1
    return Matrix(rows)

def SwapNoCopy(matrix : Matrix,row1 : list,row2 : list) -> None:
    """Swaps the rows of a matrix given a list (DOES NOT CREATE A COPY OF THE MATRIX IT THE OPERATIONS ARE PERFORMED ON THE MATRIX)"""
    rows = matrix.rawMatrix()
    is_duplicate_1 = True if rows.count(row1) > 1 else False
    is_duplicate_2 = True if rows.count(row2) > 1 else False
    if is_duplicate_1 or is_duplicate_2:
        if is_duplicate_1 and is_duplicate_2:
            duplicate = "Row 1 and Row 2 :  {},{}\n were".format(row1,row2)
        else:
            duplicate = "Row 1 : {} was".format(row1) if is_duplicate_1 else "Row 2 : {} was".format(row2)

        raise ValueError(f'{duplicate} found more than once in the Matrix.')
    #Get each index
    index_1 = rows.index(row1)
    index_2 = rows.index(row2)
    rows[index_1] = row2
    rows[index_2] = row1
    return None

def CreateMatrixPassingCollumns(array_of_arrays : list) -> Matrix:
    """
    Instead of passing the rows of the matrix,
    you pass the collumns and it creates the matrix
    [     #C1     #C2      #C3
        [1,2,3],[4,5,6],[7,8,9]


        [1,4,7],
        [2,5,8],
        [3,6,9]

    ]
    """
    counts = []
    for row in array_of_arrays:
        counts.append(len(row))
    if counts.count(counts[0]) != len(counts):
        raise ValueError("All arrays inside the base array must have equal lenght!")
    ROWS = [[] for item in range(len(array_of_arrays[0]))] 
    for col in array_of_arrays:
        i = 0
        for num in col:
            ROWS[i].append(num)
            i+=1
    return Matrix(ROWS)

def combine(r1, r2, scalar):
    r1[:] = [x-scalar*y for x,y in zip(r1,r2)]
        
def divide_by_dividor(li,scalar):
    li[:] = [x/scalar for x in li]


def ref(matrix : Matrix) -> Matrix:
    """Returns the reduced echelon form of a matrix (not RREF it does not handle upper diagnals)
       EXAMPLE : 
        Y = Matrix([
            [1,2,3],
            [4,5,6],
            [7,8,9]
        ])

        ref(Y)

        CI |   C1       C2      C3
        R1 |  1.0      1.14    1.29
        R2 |  0.0       1.0     2.0
        R3 | -0.0      -0.0     1.0

    """
    copy = Matrix([item[:] for item in matrix.rawMatrix()]) #Create a copy of the matrix
    matrix_copy = copy.rawMatrix() 
    cur_row=0  #Starting index for rows
    for j in range(0,matrix.collumns): #Iterate as much as the number of collumns
        max_element = abs(matrix_copy[cur_row][j]) #Find the max element
        pos_max = 0
        for i,x in enumerate(matrix_copy[cur_row:]):
            if abs(x[j])>max_element :
                max_element = abs(x[j])
                pos_max = i
        pos_max += cur_row
        temp = matrix_copy[pos_max]
        matrix_copy[pos_max]=matrix_copy[cur_row]
        matrix_copy[cur_row] = temp
        if matrix_copy[cur_row][j]!=0:
            divide_by_dividor(matrix_copy[cur_row],matrix_copy[cur_row][j])
            pivot = matrix_copy[cur_row]
            for i,line in enumerate(matrix_copy[cur_row+1:]):
                if line[j]!=0:
                    combine(line,pivot,line[j])
            cur_row+=1
        if cur_row==copy.__len__()[0]:
            break
    return Matrix(matrix_copy)

def isConsistent(coefficient_matrix : Matrix,augmented_matrix : Matrix) -> bool:
    """Determines wheter a system of equations is consistent"""
    r1 : int = coefficient_matrix.rank()
    r2 : int = augmented_matrix.rank()
    return r1 == r2

def SolveCramer(matrix : Matrix,unknowns : Union[tuple,list],outputs : Vector, ContinueOnFail : bool = False ) -> dict:
    """
        Solves a system of linear equations given The equations in Matrix format, the names of the unknowns and the desired outputs in a tuple
        As the name suggest it uses cramer's rule computing the determinant's of each matrix and dividing the by the determinant of the base matrix
        \nThe following system of equations:
        -----------
        2x+y-z=1
        3x+2y-2z=13
        4x-2y+3z=9
        ------------ 
        \nWould be translated into this form:

        # For each of the coefficients of the equations put them in a Matrix
        matrix = Matrix([
            [2,  1, -1],
            [3,  2,  2],
            [4, -2,  3]
        ])

        # Wrap the desired outputs into a Vector 
        outputs = Vector([1,13,9])

        #Finally wrap your unknowns in a tuple
        unknowns = ('x','y','z')

        #Apply cramer's rule
        print(SolveCramer(matrix,unknowns,outputs))

        \nThis would output the following:
        {'x': 1.0, 'y': 2.0, 'z': 3.0}

        \nif there are no solutions or there are infinetely many of them an Exception is raised 

    """
    base_determinant = matrix.determinant() #This is the determinant of the base system that always stays constant
    #If it is 0 either it has no solution or infinite
    if base_determinant==0:
        augmented_matrix = matrix.appendCollumn(outputs.getMatrix())
        assert isConsistent(matrix,augmented_matrix),"Inconsistent System : 0 Solutions (Coefficient Matrix rank != Augmented Matrix rank)"
        if not ContinueOnFail:
            raise RuntimeError("Root Calclulation Failed : Determinant is 0 and system has infinite solutions (stopped because ContinueOnFail was {})".format(ContinueOnFail))
        return matrix.solve(outputs,unknowns,useRef=True)
    raw_outputs = outputs.getMatrix()
    determinants = []
    for i in range(matrix.collumns): #3
        new_matrix = [[] for item in matrix.rawMatrix()]
        new_matrix[i].extend(raw_outputs) #Puts the outputs in the right place
        not_to_push = matrix.colls(i)
        j = 0
        for col in matrix.collsAll():
            if col != not_to_push:
                new_matrix[j].extend(col)
            j+=1
        determinants.append(determinant(CreateMatrixPassingCollumns(new_matrix)))
    variables = {}
    

    for variable in unknowns:
        variables[variable] = None
    k = 0
    for var in variables:
        variables[var] = determinants[k] /base_determinant
        k+=1   
    return variables


def solveREF(matrix : Matrix,unknowns : Union[tuple,list],Output: Vector, onFailSetConst : Tuple[int,str,float] = 1) -> dict:
    """Solves a system of linear equations using backsubtitution,
       after performing row reduction on a given matrix
       
       What to pass in :

       - You must pass the equations in Matrix format
       - you must provide the outputs in a Vector object
       - You must provide a tuple or a list of the variable names for your Unknwown
    
    \nThe following system of linear equations:
        ------------
        2x+y-z=1
        3x+2y+2z=13
        4x-2y+3z=9
        ------------
    Would be transleted into this :

    Y = Matrix([ #The basic matrix
        [2,1,-1],
        [3,2,2],
        [4,-2,3]
    ])

    unknowns = ('x','y','z') #Each of the unknowns corresponds to a certain collumn

    output = Vector([1,13,9]) #The output in Vector format
    """
    copy_matrix = Matrix([row[:] for row in matrix])
    collumns = [col[:] for col in copy_matrix.collsAll()]
    output = Output.getMatrix()
    collumns.append(output)
    final_matrix = CreateMatrixPassingCollumns(collumns) #matrix from collumns
    reduced_matrix = ref(final_matrix) #row reduction performed
    #c*z = 0.86 => z = 0.86 / c
    try:
        z = reduced_matrix.index(-1,-1) / reduced_matrix.index(-1,-2)
    except: #No Solutions or infinite
        assert isConsistent(copy_matrix,final_matrix),"Inconsistent System : 0 Solutions (Coefficient Matrix rank != Augmented Matrix rank)"
        z = onFailSetConst #NOTE : HERE IF THE MATRIX IS CONSISTENT 1 IS PICKED AS THE FINAL VALUE Z (GIVING 1 OF THE INFINITE SOLUTIONS)
    VALUES = []
    VALUES.append(z) #Begin by having the value of z inside the values
    iterable = list(reversed([row for row in reduced_matrix.rawMatrix()]))
    iterable.pop(0) #We have already found the first parameter
    for row in iterable:
        TMP_SUM = []
        #All the non-zero elements
        target = row.pop(-1)
        sub_argument = [item for item in row if item != 0]
        divisor = sub_argument.pop(0)
        l = 0
        for remaining_num in sub_argument:
            TMP_SUM.append(remaining_num * list(reversed(VALUES))[l])
            l+=1
        VALUES.append(  (target - sum(TMP_SUM)) / divisor ) #bring the constants to the other side and divide
    VALUES = list(reversed(VALUES))
    result = {}
    m = 0
    for var in unknowns:
        result[var] = VALUES[m]
        m+=1
    return result


def Vector0(dimensions : int) -> Vector:
    return Vector([0 for _ in range(dimensions)])

def rank(matrix : Matrix) -> int:
    """
    The number of non-zero rows on the reduced Matrix,
    (Checks if one row is a combination of the other ones)
    """
    row_reducted_matrix =  ref(matrix)
    __rank__ = 0
    for row in row_reducted_matrix.rawMatrix():
        if row.count(0) != len(row):
            __rank__ +=1
    return __rank__

def CharacteristicPolynomial(square_maxtirx : Matrix,returnSub : bool = False) -> pl.Polynomial:
    """
        Returns a Polynomial whose roots are the eigenvalues of the passed in Matrix,
        if returnSub it will return square_maxtirx - lamda_identity, where the lamda
        identity is the identity matrix multiplied by lamda
    """
    assert square_maxtirx.is_square(), "Non square matrix attempting to find characteristic polynomial"
    dim = square_maxtirx.__len__()[0] #dimensions
    i_dim = IdenityMatrix(dimensions=dim) #Identity matrix of dimensions N
    lamda_identity = i_dim.forEach(lambda n : n * epsilon) #Multiply lamda with Matrix
    sub = square_maxtirx - lamda_identity #Subtract from base Matrix lamda multiplied
    det = sub.determinant() #The Characteristic Polynomial
    if not returnSub:
        return det #Return only polynomial
    return det,sub #Return determinant and square_matrix - lamda_identity_matrix

def eigenValues(square_maxtirx : Matrix,iterations : int = 50):
    """Returns the eigen values of a given square Matrix"""
    char_pol = CharacteristicPolynomial(square_maxtirx)
    return char_pol.roots(iterations) #find the roots of the Polynomial

def substitute(expression : Any,term : Union[float,complex]):
    if not type(expression) == POLY:
        return expression
    return expression.getFunction()(term)

def EigenVectors(square_maxtirx : Matrix,iterations : int = 50) -> Dict[Union[complex,float],Vector]:
    char_pol = CharacteristicPolynomial(square_maxtirx,returnSub=True)
    sub = char_pol[1]
    eigen_values = char_pol[0].roots(iterations=iterations) #The roots are the eigen Values
    dim = square_maxtirx.__len__()[0] #dimensions
    unknowns = [i for i in range(dim)]
    output = Vector0(dim)
    eigen_values_vectors = {}
    for root in eigen_values:
        m_0 = sub.forEach(lambda num : substitute(num,root)) #Substitute the Eigen Values in the Lamda scaled Identity Matrix
        eigen_vector = m_0.solve(output,unknowns,useRef=True) #Solve the linear system
        eigen_vector = Vector([eigen_vector.get(sol) for sol in eigen_vector])
        eigen_values_vectors[root] = eigen_vector #Eigen value has a coressponding Vector
    return eigen_values_vectors


def magnitude(vector : Vector) -> Union[float,int]:
    components = []
    for i in range(len(vector)):
        components.append(vector[i]**2)
    return sqrt(sum(components))

def AngleBetweenVectors(vector_0 : Vector,vector_1 : Vector, degrees  : bool = False) -> Union[float,int]:
    a : Vector = vector_0
    b : Vector = vector_1
    div : float = a.dot(b) / (magnitude(a) * magnitude(b))
    return arccos(div,degrees=degrees)

if __name__ == "__main__":
    pass    


"""
    \tRandom submodule
    This is a custom made random submodule,
    that includes all trivial random operations (nothing more than the builtin),
    - random() #Pseudorandom number between 0,1
    - randendint(end : int) #Random in range (0,end)
    - choices(array : list,num_choice : int) #Gives an list of num_choice lenght containing random choices from the passed in list
    - randrange(start,end) #Returns a random floating point number in range(start,end)
    - randint(start,end) #Returns int(randrange(start,end))
    - random_complex() #Returns random() * complex(1,1)

"""
from time import time #for seed generation
from .trigonometric import sin #Acting as an 'activation' function for squishing values to (0,1)
from typing import Union

seed = time()
addition = 2.718281828459045 + 3.141592653589793
modulus = 50.12

TEMP_MEMORY = [int(str(seed)[-3:].replace(".",'1'))]

def random() -> float:
    """
        Generates a pseudo-random number between zero and one
    """
    # 0.008620870113372803 for 1k iterations
    val =  (TEMP_MEMORY[-1] + addition) % modulus
    TEMP_MEMORY[0] = val
    return abs(sin(val,iterations=10,degrees=True))

def randendint(end : int):
    """Returns a random integer in range (0,end)"""
    return round(random() * end)

def choice(array : list) -> list:
    """
    Returns a pseudo-randomly chosen element from the input array 
    """
    return array[randendint(len(array))]

def choices(array : list,num_choice : int) -> list:
    """Given an array and a variable 'num_choice' it gives back\n
       a new list containing a total of num_choice,\n
       pseudo-randomly chosen elements\n
    """
    choice_array = []
    for _ in range(num_choice):
        choice_array.append(choice(array))
    return choice_array

def randrange(start : Union[float,int],end : Union[float,int]) -> float:
    """Returns a random floating point number
       given a start and an end
    """
    return random() * (end-start) + start

def randint(start : Union[float,int],end : Union[float,int]) -> float:
    """Returns a random integer given a minimun and a maximun value"""
    return round(randrange(start,end))

def random_complex(factor : Union[float,int] = 1) -> complex:
    """Returns a random complex number scaled by a factor"""
    return random() * complex(1,1) * factor

if __name__ == "__main__": 
    pass


"""
    Computes powers and roots of specified numbers using Newton's approximation,
    it also handle operation with imaginary numbers

    Detail:
        ** Power : 
            ** Base : float or complex
            ** Exponent : float or complex
            returns base to the power of exponent
        ** sqrt :
            x : float or int
            return the square root of x
        ** nthRoot :
            A generilization of the sqrt function

"""

from .basic import product,isInteger
from . import functions
from typing import Union
from . import trigonometric as trg

def integerPow(base : float,exponent : int) -> int:
    """Used exponentiation to an integer exponent"""
    return product(*[base for i in range(int(exponent))])

def floatPow(base,exponent):
    """For handling float powers"""
    total = 1
    constLog = functions.ln(base)
    for i in reversed(range(1,50)):
        total = 1 + total * exponent * constLog / i
    return total

def power(base : Union[float,complex],exponent : Union[float,complex]) -> Union[float,complex]:
    """
    The power function equivalant of the python operation a**b\n
    it can handle any floating or integer value for a base\n
    for an exponent it can handle any integer or float or complex number ,\n
    it returns a floating point number that may not be 100% accurate,
    since it uses infinite sum approximation

    Here is how it treats complex numbers (exponents) => :
        ** e^(ix) = cos(x) + i *sin(x) #cis(x) for short
        ** e^(i*ln(a)) = cis(ln(a))
        ** e^(ln(a^i)) = cis(ln(a))
        ** a^i = cis(ln(a))
        ** (a^i)^b = (cis(ln(a)))^b
        ** a^(bi) + a^c = (cis(ln(a)))^b + a^c
        ** a^(bi+c) = (cis(ln(a)))^b + a^c
    Complex Base number is treated normally like an integer

    """
    TYPE = type(exponent)

    if TYPE == complex:
        s = exponent # a*i+b
        return power(base,s.real) + power(functions.cis(functions.ln(base)),s.imag)

    if exponent < 0:
        return 1 / power(base,-exponent)

    if TYPE in (float,int):
        return integerPow(base,exponent) if isInteger(exponent) else floatPow(base,exponent)

    raise TypeError("Values {} and {} of type {},{} do not support the pow function".format(TYPE,type(base),exponent,base))


def sqrt(x : float) -> Union[float,complex]:
    """
    Returns the square root of a given number which could be either an integer or a float,
    and reutrns either a floating number which may have some accuracy issues or a complex number,
    Does not use some special approximation technique, reutrns the power function of power(x,1/2)
    """
    return floatPow(x,0.5) if x >= 0 else complex(0,1) * sqrt(abs(x))

def nthRoot(subroot : float,n : float) -> float:
    """Does not use some special approximation technique, reutrns the power function of power(x,1/n)"""
    if n==0:
        return 1
    return floatPow(subroot,1/n)

def sqrt_subfunction(x,c):
    """Function used to estimate the sqrt"""
    return power(x,2) - c

def sqrt_newton(x : float,iterations : int = 100,catchNegativeRoot=False) -> Union[float,complex]:
    """
        Uses Newtown's method for calculating the square root of a specific number,\n
        you can specify with 'iterations' how many times you want to repeat the algorithm,\n
        if the input argument is less than 0 it will return a complex number
    """
    if x <= 0:
        if(x==0):
            return 0.0
        if catchNegativeRoot:
            raise ValueError("Value '{}' is not in the real number range".format(x))
        return complex(0,1) * sqrt(abs(x))
    point = 1
    for i in range(iterations):
        function_difference = sqrt_subfunction(point,x) / (2*point)
        point = point - function_difference
    return point

def nth_subfunction(x,exponent,c):
    """function used for the calculation of the nth root acts as a derivative"""
    return power(x,exponent) - c

def nthRootNewton(subroot : float,n : int,iterations : int = 100,catchNegativeRoot=False) -> float:
    """
        Uses Newtown's method for calculating the nth root function of a specific number,\n
        you can specify with 'iterations' how many times you want to repeat the algorithm,\n
        You can specify whether you want to throw an error when the result is complex,\n
        If not specified it will return the complex solution
    """
    if(n%2==0) or n==0:
        if n==0:
            return 0.0
        if subroot < 0:
            if not catchNegativeRoot:
                return trg.complexRoot(n) * nthRootNewton(abs(subroot),n,iterations=iterations)
            raise ValueError("Even root must contain only positive floating values not {}".format(subroot))
    
    def diffeq(x):
        raised = power(x,n-1)
        return n*raised

    point = 1
    for i in range(iterations):
        function_difference = nth_subfunction(point,n,subroot) / diffeq(point)
        point = point - function_difference

    return point

if  __name__ == "__main__":
    print(power(2,complex(2,1)))


"""
    Module containing some popular mathematical functions,\n
    that either have no specific category or were just forced here:
        ** exp(x) (Taylor Expansion or just e**x) #float -> float
        ** ln(x) #float (positive or not),complex -> float,Complex
        ** log(x) look above
        ** factorial(n) #Integer -> Integer
        ** doubleFactorial(n) #Integer -> Integer
        ** fibonacci(n) #Integer -> Integer
        ** erf(x) #float -> float
        ** erfi(x) #Complex version
        ** cis(x) # cos(x) + i*sin(x)
        ** quadratic(x) #float,int -> float,complex (linear or quadratic)
"""

from . import powers
from functools import lru_cache
from .constants import e,pi,imaginary
from .num_theory import isEven,isOdd
from typing import Union,Tuple,Optional
from . import trigonometric as trig

INFINITESIMAL = 1 / 10000 #For limit approximation (1e-5)

def compose(function_0 : callable,function_1 : callable,point : Optional[Union[float,int]]) -> Union[callable,float]:
    comp = lambda x: function_0(function_1(x))
    return comp(point) if point is not None else comp

@lru_cache(maxsize=1000)
def factorial(n : int) -> int:
    """
    Returns the product of all the numbers in range(1,n)\n
    ** Uses the built in functool's module lru_cache decorator 
    """
    if n in (1,0):
        return 1
    return n * factorial(n-1)


@lru_cache(maxsize=1000)
def doubleFactorial(n : int) -> int:
    """ if n is even it returns the sum of all the even numbers\n
        if n is odd it returns the sum of all the odd numbers\n
        in the range(1,n)\n
       ** Uses the built in functool's module lru_cache decorator 
    """
    if n in (1,0):
        return 1
    return n * doubleFactorial(n-2)

@lru_cache(maxsize=1000)
def fibonacci(n : int) -> int:
    """Returns the fibbonacci function at a specific point\n
       ** Uses the built in functool's module lru_cache decorator 
    """
    if n in (0,1):
        return 1
    return fibonacci(n-1) + fibonacci(n-2)

def exp(x : float,iterations : int = 100,taylor_exapnsion=False):
    """Calulates the exponential function,\n
        if taylor_exapnsion is set to True it will do what it says,\n
        use the taylor expansion of the exp function for calculations,\n
        else it will use the stored constant e and raise it to the power,\n
        if you set taylor_exapnsion=True you can specify how many times to iterate \n
    """
    if(not taylor_exapnsion):
        return powers.power(e,x)
    return sum([powers.power(x,n) / factorial(n) for n in range(iterations)])


def ln(x : float,iterations : int = 100) -> Union[float,complex]:
    """
        Natural log function (log with base the constant e)
        it can handle either a  floating point or an imaginary number
        it uses 'infinite' sumations which you can specify the iterations
        This is the exact formula for the natural log : https://wikimedia.org/api/rest_v1/media/math/render/svg/1d9729501b26eb85764942cb112cc9885b1a6cca
        
        Here is how it handles negative values :  (log(negative) = πi + ln(abs(negative)) )
        \n\t=> e**(iπ) = -1
        \n\t=> iπ*ln(e) = ln(-1)
        \n\t=> πi = ln(-1) 
        Now with the help of this rule (log(ab) = log(a) + log(b)):
          => log(negative) = πi + ln(abs(negative)) 
          # ln(-5) = ln(-1 * 5)
          # ln(-5) = ln(-1) + ln(-5)
    """
    if type(x) in (float,int):
        total = 0
        # k 2*k+1 is always an integer
        for k in range(iterations):
            denominator = 1 / (2*k+1)
            apr = (x - 1) / (x + 1)
            final = denominator * pow(apr,2*k+1)
            total += final

        return 2*total

    if type(x) == complex or x < 0:
        if type(x) != complex:
            return (imaginary * pi) + ln(abs(x),iterations=iterations)
        real = x.real
        imag = x.imag
        suma = pow(real,2) + pow(imag,2)
        reduced_log = ln(suma) / 2
        inverseTan = complex(0,1) * trig.arctan(imag / real)
        return reduced_log + inverseTan

    raise TypeError("Logarithmic function cannot be evaluated at {}".format(x))
    
def log(of_num : float,base : float = 10) -> float:
    """
        Returns the logarithm of a number given a base (if none is proveded it defaults to 10)
        \nFor calculations it uses the following property of logs : log(a,b) = ln(a) / ln(b)
        \nThe 'of_num' parameter can also be a complex number (check the ln for more info)
    """
    return ln(of_num) / ln(base)

def erf(x : float) -> float:
    """Calculates the error function at a specific point"""
    MULTIPLIER = 2 / powers.sqrt(pi)
    total = 0
    for n in range(100):
        denominator = factorial(n) * (2*n+1)
        nominator = powers.power(-1,n) * powers.integerPow(x,2*n+1)
        total += nominator / denominator
    return MULTIPLIER * total

def erfi(x : float) -> float:
    """Calculates  the imaginary error function at a specific point"""
    MULTIPLIER = 2 / powers.sqrt(pi)
    total = 0
    for n in range(100):
        denominator = factorial(n) * (2*n+1)
        nominator = powers.integerPow(x,2*n+1)
        total += nominator / denominator
    return MULTIPLIER * total

def quadratic(a,b,c) -> Union[Tuple[complex],Tuple[float]]:
    """
        Gives all complex roots of a qudratic equations,
        if the equation is linear (a==0) it will return 1 real root,
        if it is qudratic it will return a tuple of the 2 anwsers,
        which are either both going to be floats or complex
    """
    if a == 0:
        return -c / b
    descriminant = powers.power(b,2) - 4*a*c
    if descriminant < 0 :
        descriminant = imaginary * powers.sqrt(abs(descriminant))
    r_0 = (-b + descriminant) / 2*a
    r_1 = (-b - descriminant) / 2*a
    return (r_0,r_1)

def cis(x : float) -> complex:
    """Returns the following operation : \n
        cos(x) + complex(0,1) * sin(x)
    """
    return trig.sin(x) + imaginary * trig.cos(x)

def derivative(function : callable,point : float,h=INFINITESIMAL) -> float:
    """Find the derivative a function at a specific point using the definition"""
    nominator = function(point + h) - function(point)
    denominator = h
    result = nominator / denominator
    return result

@lru_cache(maxsize=200)
def nthDerivative(f : callable,x : float,order : int) -> float:
    """Returns the nth order derivative a given function f at a given point x\n
       The order must be a positive integer greater or equal to one, else it's\n
       going to cause a recursive infinite loop,\n
       NOTE : No checking is done to ensure that order >= 1\n
    """
    h = INFINITESIMAL
    if order == 1:
        return derivative(f,x)
    return ( nthDerivative(f,x+h,order-1) - nthDerivative(f,x,order-1) ) / h

def integral(f : callable,x : Union[float,int],approximation : float = 0.01,target : float = 0.0000001) -> float:
    """
        Finds the integral of a given function at a specific point using sum approximation
    """
    difference_in_x = approximation
    start = target
    total = 0
    while True:
        if not start <= x:
            break
        total += f(start)*difference_in_x
        start += difference_in_x
    return total

    
def CurveArea(function : callable,point_0 : float,point_1 : float) -> float:
    if point_0 > point_1:
        raise ValueError("Point 0 must be less than point 1")
    return integral(function,point_1) - integral(function,point_0)

#Methods for computing roots

def NewtonMethod(function : callable,starting_point : Union[float,int],iterations : int) -> float:
    """Newton's method for root aproximation"""
    x = starting_point
    for _ in range(iterations):
        f_div = function(x) / derivative(function,x)
        x -= f_div
    return x

def SecantMethod(function : callable,starting_point_0 : Union[float,int],starting_point_1 : Union[float,int],iterations : int) -> float:
    """Secant method for root aproximation works great at low iterations"""
    #Requires 2 starting points
    x_1 = starting_point_0
    x_2 = starting_point_1
    for _ in range(iterations):
        f_1 = function(x_1) #function at x_1
        f_2 = function(x_2) #function at x_2
        nominator = x_2 * f_1 - x_1 * f_2 
        den = f_1 - f_2
        nom_den = nominator / den
        x_2 = x_1 #Set x_2 to x_1
        x_1 = nom_den #Set x_1 to the result
    return x_1

if __name__ == "__main__":
    pass